# Copyright 2019 Kemal Kurniawan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from datetime import timedelta
from typing import Any, Callable, Dict, Iterable, List
import time
import logging

from rnnr.event import Event

Handler = Callable[[dict], None]
logger = logging.getLogger(__name__)


class Runner:
    """A neural network runner.

    A runner provides a thin abstraction of iterating over batches for several epochs,
    which is typically done in neural network training. To customize the behavior during
    a run, a runner provides a way to listen to events emitted during such run.
    To listen to an event, call `Runner.append_handler` and provide the event handler.
    A handler is a callable that accepts a `dict` and returns nothing. The `dict` is
    the state of the run. By default, the state contains:

    * ``runner`` - The runner object itself.
    * ``batches`` - iterable of batches which constitutes an epoch.
    * ``max_epoch`` - maximum number of epochs to run.
    * ``epoch`` - current number of epoch. Not available to handlers of `Event.STARTED`
      and `Event.FINISHED`.
    * ``batch`` - current batch retrieved from ``state['batches']``. Only available to
      handlers of `Event.BATCH_STARTED` and `Event.BATCH_FINISHED`, as well as ``batch_fn``
      passed to `~Runner.run`.
    """

    def __init__(self) -> None:
        self._handlers: Dict[Event, List[Handler]] = defaultdict(list)
        self._running = False
        self._epoch_start_time = 0.

        self.append_handler(Event.EPOCH_STARTED, self._print_start_epoch)
        self.append_handler(Event.EPOCH_FINISHED, self._print_finish_epoch)

    def append_handler(self, event: Event, handler: Handler) -> None:
        """Append a handler for the given event.

        Args:
            event: Event to handle.
            handler: Handler for the event.
        """
        self._handlers[event].append(handler)

    def _print_start_epoch(self, state: dict) -> None:
        if state['max_epoch'] > 1:
            self._epoch_start_time = time.time()
            logger.info('Starting epoch %d/%d', state['epoch'], state['max_epoch'])

    def _print_finish_epoch(self, state: dict) -> None:
        if state['max_epoch'] > 1:
            elapsed = timedelta(seconds=time.time() - self._epoch_start_time)
            logger.info('Epoch %d/%d done in %s', state['epoch'], state['max_epoch'], elapsed)

    def on(self, event: Event) -> Callable[[Handler], Handler]:
        def decorator(handler: Handler) -> Handler:
            self.append_handler(event, handler)
            return handler

        return decorator

    def run(
            self,
            batch_fn: Callable[[dict], None],
            batches: Iterable[Any],
            max_epoch: int = 1,
    ) -> dict:
        """Run on the given batches for a number of epochs.

        Args:
            batch_fn: Function to call for each batch. This function should accept
                the state dict and may write to the state as well.
            batches: Batches to iterate over in an epoch.
            max_epoch: Maximum number of epochs to run.

        Returns:
            State of the run at the end.
        """
        self._running = True
        state: dict = {'runner': self, 'max_epoch': max_epoch, 'batches': batches}

        self._emit(Event.STARTED, state)
        for epoch in range(1, max_epoch + 1):
            if not self._running:
                break

            state['epoch'] = epoch
            self._emit(Event.EPOCH_STARTED, state)

            for batch in batches:
                if not self._running:
                    break
                state['batch'] = batch
                self._emit(Event.BATCH_STARTED, state)
                batch_fn(state)
                self._emit(Event.BATCH_FINISHED, state)

            state.pop('batch', None)
            self._emit(Event.EPOCH_FINISHED, state)
        state.pop('epoch', None)
        self._emit(Event.FINISHED, state)
        return state

    def _emit(self, event: Event, state: dict) -> None:
        for handler in self._handlers[event]:
            handler(state)

    def stop(self) -> None:
        """Stop the runner immediately after the current batch is finished.

        Note that the appropriate handlers for ``Event.*_FINISHED`` events are still called
        before the run truly stops.
        """
        self._running = False



from unittest.mock import MagicMock, Mock, call

from tqdm import tqdm

from rnnr.attachments import ProgressBar


def test_ok(runner):
    batches = range(10)
    batch_fn = lambda state: state['batch']**2
    mock_tqdm_cls = MagicMock(spec=tqdm)

    ProgressBar(tqdm_cls=mock_tqdm_cls).attach_on(runner)
    runner.run(batch_fn, batches)

    mock_tqdm_cls.assert_called_once_with(batches)
    assert not mock_tqdm_cls.return_value.set_postfix.called
    assert mock_tqdm_cls.return_value.update.mock_calls == [call(1) for b in batches]
    mock_tqdm_cls.return_value.close.assert_called_once_with()


def test_default_size_key(runner):
    batches = [list('foo'), list('quux')]
    mock_tqdm_cls = MagicMock(spec=tqdm)

    def batch_fn(state):
        state['n_items'] = len(state['batch'])

    pbar = ProgressBar(tqdm_cls=mock_tqdm_cls)
    pbar.attach_on(runner)
    runner.run(batch_fn, batches)

    assert mock_tqdm_cls.return_value.update.mock_calls == [call(len(b)) for b in batches]


def test_size_key(runner):
    batches = [list('foo'), list('quux')]
    mock_tqdm_cls = MagicMock(spec=tqdm)

    def batch_fn(state):
        state['foo'] = len(state['batch'])

    pbar = ProgressBar(tqdm_cls=mock_tqdm_cls, size_key='foo')
    pbar.attach_on(runner)
    runner.run(batch_fn, batches)

    assert mock_tqdm_cls.return_value.update.mock_calls == [call(len(b)) for b in batches]


def test_stats_key(runner):
    batches = range(10)
    mock_tqdm_cls = MagicMock(spec=tqdm)

    def batch_fn(state):
        state['stats'] = {'loss': state['batch']**2}

    pbar = ProgressBar(tqdm_cls=mock_tqdm_cls, stats_key='stats')
    pbar.attach_on(runner)
    runner.run(batch_fn, batches)

    assert mock_tqdm_cls.return_value.set_postfix.mock_calls == [
        call(loss=b**2) for b in batches
    ]


def test_with_kwargs(runner):
    batches = range(10)
    mock_tqdm_cls = MagicMock(spec=tqdm)
    kwargs = {'foo': 'bar', 'baz': 'quux'}

    ProgressBar(tqdm_cls=mock_tqdm_cls, **kwargs).attach_on(runner)
    runner.run(Mock(), batches)

    mock_tqdm_cls.assert_called_once_with(batches, **kwargs)



# Copyright 2019 Kemal Kurniawan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque
from typing import Any, Callable, Deque, Iterable, Optional, Mapping
from pathlib import Path
import logging
import pickle

logger = logging.getLogger(__name__)


class EarlyStopper:
    """A handler for early stopping.

    This handler keeps track the number of times the loss value does not improve. If this
    number is greater than the given patience, this handler stops the given runner.

    Example:

        >>> valid_losses = [0.1, 0.2, 0.3]  # simulate validation batch losses
        >>> batches = range(10)
        >>> batch_fn = lambda _: None
        >>>
        >>> from rnnr import Event, Runner
        >>> from rnnr.attachments import MeanReducer
        >>> from rnnr.handlers import EarlyStopper
        >>>
        >>> trainer = Runner()
        >>> @trainer.on(Event.EPOCH_STARTED)
        ... def print_epoch(state):
        ...     print('Epoch', state['epoch'], 'started')
        ...
        >>> @trainer.on(Event.EPOCH_FINISHED)
        ... def eval_on_valid(state):
        ...     def eval_fn(state):
        ...         state['output'] = state['batch']
        ...     evaluator = Runner()
        ...     MeanReducer(name='mean').attach_on(evaluator)
        ...     eval_state = evaluator.run(eval_fn, valid_losses)
        ...     state['loss'] = eval_state['mean']
        ...
        >>> trainer.append_handler(Event.EPOCH_FINISHED, EarlyStopper(patience=2))
        >>> _ = trainer.run(batch_fn, batches, max_epoch=7)
        Epoch 1 started
        Epoch 2 started
        Epoch 3 started
        Epoch 4 started

    Args:
        patience: Number of times to wait for the loss to improve before stopping.
        loss_key: Get the loss value from ``state[loss_key]``.
        eps: An improvement is considered only when the loss value decreases by at least
            this amount.
    """
    def __init__(
            self,
            patience: int = 5,
            loss_key: str = 'loss',
            eps: float = 1e-4,
    ) -> None:
        self._patience = patience
        self._loss_key = loss_key
        self._eps = eps

        self.min_loss = float('inf')
        self._n_bad_losses = 0

    def __call__(self, state: dict) -> None:
        loss = state[self._loss_key]
        if loss <= self.min_loss - self._eps:
            self.min_loss = loss
            self._n_bad_losses = 0
        else:
            self._n_bad_losses += 1

        if self._n_bad_losses > self._patience:
            logger.info('Patience exceeded, stopping early')
            state['runner'].stop()


class Checkpointer:
    """A handler for checkpointing.

    Checkpointing here means saving some objects (e.g., models) periodically during a run.

    Example:

        >>> from pathlib import Path
        >>> from pprint import pprint
        >>> from rnnr import Event, Runner
        >>> from rnnr.handlers import Checkpointer
        >>>
        >>> batches = range(3)
        >>> batch_fn = lambda _: None
        >>> tmp_dir = Path('/tmp')
        >>> runner = Runner()
        >>> @runner.on(Event.EPOCH_FINISHED)
        ... def store_checkpoint(state):
        ...     state['checkpoint'] = {'model.pkl': 'MODEL', 'optimizer.pkl': 'OPTIMIZER'}
        ...
        >>> runner.append_handler(Event.EPOCH_FINISHED, Checkpointer(tmp_dir, max_saved=3))
        >>> _ = runner.run(batch_fn, batches, max_epoch=7)
        >>> pprint(sorted(list(tmp_dir.glob('*.pkl'))))
        [PosixPath('/tmp/5_model.pkl'),
         PosixPath('/tmp/5_optimizer.pkl'),
         PosixPath('/tmp/6_model.pkl'),
         PosixPath('/tmp/6_optimizer.pkl'),
         PosixPath('/tmp/7_model.pkl'),
         PosixPath('/tmp/7_optimizer.pkl')]

    Args:
        save_dir: Save checkpoints in this directory.
        checkpoint_key: Get the checkpoint from ``state[checkpoint_key]``. A checkpoint
            is a mapping whose keys are filenames and the values are the objects to checkpoint.
            The filenames in this dictionary's keys are prepended with the number of times
            this handler is called to get the actual saved files' names. This allows the
            actual filenames contain the e.g. epoch number if this handler is invoked at the
            end of each epoch.
        max_saved: Maximum number of checkpoints saved.
        loss_key: Get the loss value from ``state[loss_key]``. Checkpoints are saved only
            when the loss is smaller than the minimum loss observed so far. The default
            of ``None`` means checkpoints are saved whenever this handler is called.
        save_fn: Function to invoke to save the checkpoints. If given, this must be a callable
            accepting two arguments: an object to save and a path to save it to. The default
            is to save the object using `pickle`.
        eps: The loss value must be smaller at least by this value to be considered as an
            improvement. Only used if ``loss_fn`` is given.
    """
    def __init__(
            self,
            save_dir: Path,
            checkpoint_key: str = 'checkpoint',
            max_saved: int = 1,
            loss_key: Optional[str] = None,
            save_fn: Optional[Callable[[Any, Path], None]] = None,
            eps: float = 1e-4,
    ) -> None:
        if save_fn is None:
            save_fn = self._default_save_fn

        self._save_dir = save_dir
        self._checkpoint_key = checkpoint_key
        self._max_saved = max_saved
        self._loss_key = loss_key
        self._save_fn = save_fn
        self._eps = eps

        self.min_loss = float('inf')
        self._n_calls = 0
        self._deque: Deque[int] = deque()

    @staticmethod
    def _default_save_fn(obj: Any, path: Path) -> None:
        with open(path, 'wb') as f:
            pickle.dump(obj, f)

    def __call__(self, state: dict) -> None:
        self._n_calls += 1
        ckpt = state[self._checkpoint_key]
        if self._should_save(state):
            self._save(ckpt)
        if self._should_delete():
            self._delete(ckpt.keys())

        assert self._n_saved <= self._max_saved

    @property
    def _n_saved(self) -> int:
        return len(self._deque)

    def _should_save(self, state: dict) -> bool:
        if self._loss_key is None:
            return True

        loss = state[self._loss_key]
        if loss <= self.min_loss - self._eps:
            logger.info('Found new best loss of %f', loss)
            self.min_loss = loss
            return True

        return False

    def _should_delete(self) -> bool:
        return self._n_saved > self._max_saved

    def _save(self, ckpt: Mapping[str, Any]) -> None:
        self._deque.append(self._n_calls)
        for name, obj in ckpt.items():
            path = self._save_dir / f'{self._n_calls}_{name}'
            logger.info('Saving to %s', path)
            self._save_fn(obj, path)

    def _delete(self, filenames: Iterable[str]) -> None:
        num = self._deque.popleft()
        for name in filenames:
            path = self._save_dir / f'{num}_{name}'
            if path.exists():
                path.unlink()



import statistics as stat

import pytest

from rnnr import Event
from rnnr.attachments import MeanReducer


def test_ok(runner):
    batches = range(5)
    values = [12, 7, 8, 44, -13]

    def batch_fn(state):
        state['output'] = values[state['batch']]

    r = MeanReducer()
    r.attach_on(runner)
    state = runner.run(batch_fn, batches)

    assert r.name == 'mean'
    assert state['mean'] == pytest.approx(stat.mean(values))


def test_more_than_one_epoch(runner):
    batches = range(5)
    values = [12, 7, 8, 44, -13, 78, 55, -109, 34, 10]

    def batch_fn(state):
        ix = (state['epoch'] - 1) * len(batches) + state['batch']
        state['output'] = values[ix]

    def efhandler(state):
        assert state[r.name] == pytest.approx(
            stat.mean(values[(state['epoch'] - 1) * len(batches) + b] for b in batches))

    r = MeanReducer()
    r.attach_on(runner)
    runner.append_handler(Event.EPOCH_FINISHED, efhandler)
    runner.run(batch_fn, batches, max_epoch=2)


def test_custom_name(runner):
    r = MeanReducer(name='avg')
    assert r.name == 'avg'


def test_value_key(runner):
    batches = range(10)

    def batch_fn(state):
        state['value'] = state['batch']**3

    r = MeanReducer(value_key='value')
    r.attach_on(runner)
    state = runner.run(batch_fn, batches)

    assert state[r.name] == pytest.approx(stat.mean(b**3 for b in batches))


def test_size_key(runner):
    batches = range(5)
    sizes = [3, 4, 9, 10, 2]

    def batch_fn(state):
        state['output'] = state['batch']
        state['foo'] = sizes[state['batch']]

    r = MeanReducer(size_key='foo')
    r.attach_on(runner)
    state = runner.run(batch_fn, batches)

    assert state[r.name] == pytest.approx(sum(batches) / sum(sizes))



# Copyright 2019 Kemal Kurniawan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum, auto


class Event(Enum):
    """An enumeration of events.

    Attributes:
        STARTED: Emitted once at the start of a run.
        EPOCH_STARTED: Emitted at the start of each epoch.
        BATCH_STARTED: Emitted at the start of each batch.
        BATCH_FINISHED: Emitted every time a batch is finished.
        EPOCH_FINISHED: Emitted every time an epoch is finished.
        FINISHED: Emitted once when a run is finished.
    """
    STARTED = auto()
    EPOCH_STARTED = auto()
    BATCH_STARTED = auto()
    BATCH_FINISHED = auto()
    EPOCH_FINISHED = auto()
    FINISHED = auto()



from unittest.mock import Mock, call
import pickle

import pytest

from rnnr.handlers import Checkpointer


def test_ok(tmp_path):
    n_calls = 5
    objs_values = {
        'model.pkl': [f'MODEL_{i}' for i in range(n_calls)],
        'opt.pkl': [f'OPT_{i}' for i in range(n_calls)],
    }

    ckptr = Checkpointer(tmp_path)
    for i in range(n_calls):
        ckpt = {name: values[i] for name, values in objs_values.items()}
        ckptr({'checkpoint': ckpt})

    for name in objs_values:
        path = tmp_path / f'{n_calls}_{name}'
        assert path.exists()
        with open(path, 'rb') as f:
            assert pickle.load(f) == objs_values[name][-1]


def test_max_saved(tmp_path):
    n_calls, max_saved = 5, 3
    objs_values = {
        'model.pkl': [f'MODEL_{i}' for i in range(n_calls)],
        'opt.pkl': [f'OPT_{i}' for i in range(n_calls)],
    }

    ckptr = Checkpointer(tmp_path, max_saved=max_saved)
    for i in range(n_calls):
        ckpt = {name: values[i] for name, values in objs_values.items()}
        ckptr({'checkpoint': ckpt})

    for name in objs_values:
        assert len(list(tmp_path.glob(f'*_{name}'))) == max_saved
        for i in range(max_saved):
            path = tmp_path / f'{n_calls - i}_{name}'
            assert path.exists()
            with open(path, 'rb') as f:
                assert pickle.load(f) == objs_values[name][n_calls - i - 1]


def test_loss_key(tmp_path):
    n_calls, max_saved = 5, 2
    objs_values = {
        'model.pkl': [f'MODEL_{i}' for i in range(n_calls)],
        'opt.pkl': [f'OPT_{i}' for i in range(n_calls)],
    }
    # new best losses are call #1, #3, and #5
    losses = [3, 4, 2, 4, 1]
    min_losses = [3, 3, 2, 2, 1]

    ckptr = Checkpointer(tmp_path, max_saved=max_saved, loss_key='loss')
    for i in range(n_calls):
        ckpt = {name: values[i] for name, values in objs_values.items()}
        ckptr({'loss': losses[i], 'checkpoint': ckpt})
        assert ckptr.min_loss == pytest.approx(min_losses[i])

    # saved call are #3 and #5 (the last 2)
    saved_calls = [3, 5]
    for name in objs_values:
        assert len(list(tmp_path.glob(f'*_{name}'))) == max_saved
        for c in saved_calls:
            path = tmp_path / f'{c}_{name}'
            assert path.exists()
            with open(path, 'rb') as f:
                assert pickle.load(f) == objs_values[name][c - 1]


def test_checkpoint_key(tmp_path):
    n_calls = 5
    objs_values = {
        'model.pkl': [f'MODEL_{i}' for i in range(n_calls)],
        'opt.pkl': [f'OPT_{i}' for i in range(n_calls)],
    }

    ckptr = Checkpointer(tmp_path, checkpoint_key='foo')
    for i in range(n_calls):
        ckpt = {name: values[i] for name, values in objs_values.items()}
        ckptr({'foo': ckpt})

    for name in objs_values:
        path = tmp_path / f'{n_calls}_{name}'
        assert path.exists()
        with open(path, 'rb') as f:
            assert pickle.load(f) == objs_values[name][-1]


def test_save_fn(tmp_path):
    n_calls = 5
    objs_values = {
        'model.pkl': [f'MODEL_{i}' for i in range(n_calls)],
        'opt.pkl': [f'OPT_{i}' for i in range(n_calls)],
    }
    mock_save_fn = Mock()

    ckptr = Checkpointer(tmp_path, save_fn=mock_save_fn)
    for i in range(n_calls):
        ckpt = {name: values[i] for name, values in objs_values.items()}
        ckptr({'checkpoint': ckpt})

    assert mock_save_fn.mock_calls == [
        call(objs_values[name][i], tmp_path / f'{i+1}_{name}')
        for i in range(n_calls)
        for name in objs_values
    ]



from functools import reduce

from rnnr.attachments import LambdaReducer


def test_ok(runner):
    outputs = [4, 2, 1, 5, 6]
    batches = range(len(outputs))

    def batch_fn(state):
        state['output'] = outputs[state['batch']]

    r = LambdaReducer('product', lambda x, y: x * y)
    r.attach_on(runner)
    state = runner.run(batch_fn, batches)

    assert r.name == 'product'
    assert state['product'] == reduce(lambda x, y: x * y, outputs)


def test_value_key(runner):
    outputs = [4, 2, 1, 5, 6]
    batches = range(len(outputs))

    def batch_fn(state):
        state['value'] = outputs[state['batch']]

    r = LambdaReducer('product', lambda x, y: x * y, value_key='value')
    r.attach_on(runner)
    state = runner.run(batch_fn, batches)

    assert state['product'] == reduce(lambda x, y: x * y, outputs)



from unittest.mock import Mock, patch

from rnnr import Event


class TestRun:
    def test_ok(self, runner):
        batches, n_calls = range(10), 0

        def batch_fn(state):
            nonlocal n_calls
            assert state['runner'] is runner
            assert state['batches'] == batches
            assert state['max_epoch'] == 1
            assert state['epoch'] == 1
            assert state['batch'] == batches[n_calls]
            n_calls += 1

        state = runner.run(batch_fn, batches)

        assert state['batches'] == batches
        assert state['max_epoch'] == 1

    def test_more_than_one_epoch(self, runner):
        batches, max_epoch, n_calls = range(10), 5, 0

        def batch_fn(state):
            nonlocal n_calls
            assert state['max_epoch'] == max_epoch
            assert state['epoch'] == n_calls // len(batches) + 1
            assert state['batch'] == batches[n_calls % len(batches)]
            n_calls += 1

        runner.run(batch_fn, batches, max_epoch=max_epoch)


class TestAppendHandler:
    def test_started(self, runner):
        batches, max_epoch = range(10), 5

        def on_started(state):
            assert set(state) == {'runner', 'batches', 'max_epoch'}
            assert state['runner'] is runner
            assert state['batches'] == batches
            assert state['max_epoch'] == max_epoch

        runner.append_handler(Event.STARTED, on_started)
        runner.run(Mock(), batches, max_epoch=max_epoch)

    def test_epoch_started(self, runner):
        batches, max_epoch, n_calls = range(10), 5, 0

        def on_epoch_started(state):
            nonlocal n_calls
            assert set(state) == {'runner', 'batches', 'max_epoch', 'epoch'}
            assert state['runner'] is runner
            assert state['batches'] == batches
            assert state['max_epoch'] == max_epoch
            assert state['epoch'] == n_calls + 1
            n_calls += 1

        runner.append_handler(Event.EPOCH_STARTED, on_epoch_started)
        runner.run(Mock(), batches, max_epoch=max_epoch)

    def test_batch_started(self, runner):
        batches, max_epoch, n_calls = range(10), 5, 0

        def on_batch_started(state):
            nonlocal n_calls
            assert set(state) == {'runner', 'batches', 'max_epoch', 'epoch', 'batch'}
            assert state['runner'] is runner
            assert state['batches'] == batches
            assert state['max_epoch'] == max_epoch
            assert state['epoch'] == n_calls // len(batches) + 1
            assert state['batch'] == batches[n_calls % len(batches)]
            n_calls += 1

        runner.append_handler(Event.BATCH_STARTED, on_batch_started)
        runner.run(Mock(), batches, max_epoch=max_epoch)

    def test_batch_finished(self, runner):
        batches, max_epoch, n_calls = range(10), 5, 0

        def batch_fn(state):
            state['output'] = state['batch']**2

        def on_batch_finished(state):
            nonlocal n_calls
            assert set(state) == {'runner', 'batches', 'max_epoch', 'epoch', 'batch', 'output'}
            assert state['runner'] is runner
            assert state['batches'] == batches
            assert state['max_epoch'] == max_epoch
            assert state['epoch'] == n_calls // len(batches) + 1
            assert state['batch'] == batches[n_calls % len(batches)]
            assert state['output'] == state['batch']**2
            n_calls += 1

        runner.append_handler(Event.BATCH_FINISHED, on_batch_finished)
        runner.run(batch_fn, batches, max_epoch=max_epoch)

    def test_epoch_finished(self, runner):
        batches, max_epoch, n_calls = range(10), 5, 0

        def on_epoch_finished(state):
            nonlocal n_calls
            assert set(state) == {'runner', 'batches', 'max_epoch', 'epoch'}
            assert state['runner'] is runner
            assert state['batches'] == batches
            assert state['max_epoch'] == max_epoch
            assert state['epoch'] == n_calls + 1
            n_calls += 1

        runner.append_handler(Event.EPOCH_FINISHED, on_epoch_finished)
        runner.run(Mock(), batches, max_epoch=max_epoch)

    def test_finished(self, runner):
        batches, max_epoch = range(10), 5

        def on_finished(state):
            assert set(state) == {'runner', 'batches', 'max_epoch'}
            assert state['runner'] is runner
            assert state['batches'] == batches
            assert state['max_epoch'] == max_epoch

        runner.append_handler(Event.FINISHED, on_finished)
        runner.run(Mock(), batches, max_epoch=max_epoch)


class TestStop:
    def test_on_batch_started(self, runner):
        mock_eshandler = Mock()
        mock_bfhandler = Mock()
        mock_efhandler = Mock()
        batches = range(10)

        def bshandler(state):
            if state['batch'] == 3:
                runner.stop()

        runner.append_handler(Event.EPOCH_STARTED, mock_eshandler)
        runner.append_handler(Event.BATCH_STARTED, bshandler)
        runner.append_handler(Event.BATCH_FINISHED, mock_bfhandler)
        runner.append_handler(Event.EPOCH_FINISHED, mock_efhandler)
        runner.run(Mock(), batches, max_epoch=2)

        assert mock_eshandler.call_count == 1
        assert mock_bfhandler.call_count == 4
        assert mock_efhandler.call_count == 1

    def test_on_epoch_started(self, runner):
        mock_bshandler = Mock()
        mock_bfhandler = Mock()
        mock_efhandler = Mock()
        batches = range(10)

        def eshandler(state):
            if state['epoch'] == 1:
                runner.stop()

        runner.append_handler(Event.EPOCH_STARTED, eshandler)
        runner.append_handler(Event.BATCH_STARTED, mock_bshandler)
        runner.append_handler(Event.BATCH_FINISHED, mock_bfhandler)
        runner.append_handler(Event.EPOCH_FINISHED, mock_efhandler)
        runner.run(Mock(), batches, max_epoch=7)

        assert mock_bshandler.call_count == 0
        assert mock_bfhandler.call_count == 0
        assert mock_efhandler.call_count == 1


def test_on_decorator(runner):
    with patch.object(runner, 'append_handler', autospec=True) as mock_append_handler:

        @runner.on(Event.BATCH_STARTED)
        def handler(state):
            pass

        mock_append_handler.assert_called_once_with(Event.BATCH_STARTED, handler)



# Copyright 2019 Kemal Kurniawan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Optional, Type
import abc

from tqdm import tqdm

from .event import Event
from .runner import Runner


class Attachment(abc.ABC):
    """An abstract base class for an attachment."""
    @abc.abstractmethod
    def attach_on(self, runner: Runner) -> None:
        """Attach to the given runner.

        Args:
            runner: Runner to attach to.
        """
        pass


class ProgressBar(Attachment):
    """An attachment to display a progress bar.

    The progress bar is implemented using `tqdm`_.

    Example:

        >>> from rnnr import Runner
        >>> from rnnr.attachments import ProgressBar
        >>> runner = Runner()
        >>> ProgressBar().attach_on(runner)
        >>> _ = runner.run(lambda _: None, range(10), max_epoch=10)

    Args:
        size_key: Get the size of a batch from ``state[size_key]`` to update the
            progress bar with. If not given, the default is to always set 1 as the size of
            a batch.
        stats_key: Get the batch statistics from ``state[stats_key]`` and display it along
            with the progress bar. The statistics dictionary has the names of the statistics
            as keys and the statistics as values.
        **kwargs: Keyword arguments to be passed to `tqdm`_ class.


    .. _tqdm: https://github.com/tqdm/tqdm
    """
    def __init__(
            self,
            size_key: str = 'n_items',
            stats_key: Optional[str] = None,
            tqdm_cls: Optional[Type[tqdm]] = None,
            **kwargs,
    ) -> None:
        if tqdm_cls is None:  # pragma: no cover
            tqdm_cls = tqdm

        self._tqdm_cls = tqdm_cls
        self._size_key = size_key
        self._stats_key = stats_key
        self._kwargs = kwargs

        self._pbar: tqdm

    def attach_on(self, runner: Runner) -> None:
        runner.append_handler(Event.EPOCH_STARTED, self._create)
        runner.append_handler(Event.BATCH_FINISHED, self._update)
        runner.append_handler(Event.EPOCH_FINISHED, self._close)

    def _create(self, state: dict) -> None:
        self._pbar = self._tqdm_cls(state['batches'], **self._kwargs)

    def _update(self, state: dict) -> None:
        if self._stats_key is not None:
            self._pbar.set_postfix(**state[self._stats_key])
        self._pbar.update(state.get(self._size_key, 1))

    def _close(self, state: dict) -> None:
        self._pbar.close()


class LambdaReducer(Attachment):
    """An attachment to compute a reduction over batches.

    This attachment gets the value of each batch and compute a reduction over them
    at the end of each epoch.

    Example:

        >>> from rnnr import Runner
        >>> from rnnr.attachments import LambdaReducer
        >>> runner = Runner()
        >>> LambdaReducer('product', lambda x, y: x * y).attach_on(runner)
        >>> def batch_fn(state):
        ...     state['output'] = state['batch']
        ...
        >>> state = runner.run(batch_fn, [10, 20, 30])
        >>> state['product']
        6000

    Args:
        name: Name of this attachment to be used as the key in the runner's
            state dict to store the reduction result.
        reduce_fn: Reduction function. It should accept two batch values and
            return their reduction result.
        value_key: Get the value of a batch from ``state[value_key]``.
    """
    def __init__(
            self,
            name: str,
            reduce_fn: Callable[[Any, Any], Any],
            value_key: str = 'output',
    ) -> None:
        self.name = name
        self._reduce_fn = reduce_fn
        self._value_key = value_key

    def attach_on(self, runner: Runner) -> None:
        runner.append_handler(Event.EPOCH_STARTED, self._reset)
        runner.append_handler(Event.BATCH_FINISHED, self._update)
        runner.append_handler(Event.EPOCH_FINISHED, self._compute)

    def _reset(self, state: dict) -> None:
        self._result = None

    def _update(self, state: dict) -> None:
        if self._result is None:
            self._result = state[self._value_key]
        else:
            self._result = self._reduce_fn(self._result, state[self._value_key])

    def _compute(self, state: dict) -> None:
        state[self.name] = self._result


class MeanReducer(LambdaReducer):
    """An attachment to compute a mean over batch statistics.

    This attachment gets the value from each batch and compute their mean at the end of
    every epoch.

    Example:

        >>> from rnnr import Runner
        >>> from rnnr.attachments import MeanReducer
        >>> runner = Runner()
        >>> MeanReducer().attach_on(runner)
        >>> def batch_fn(state):
        ...     state['output'] = state['batch']
        ...
        >>> state = runner.run(batch_fn, [1, 2, 3])
        >>> state['mean']
        2.0

    Args:
        name: Name of this attachment to be used as the key in the runner's state
            dict to store the mean value.
        value_key: Get the value of a batch from ``state[value_key]``.
        size_key: Get the size of a batch from ``state[size_key]``. If
            the state has no such key, the size defaults to 1. The sum of all these
            batch sizes is the divisor when computing the mean.
    """
    def __init__(
            self,
            name: str = 'mean',
            value_key: str = 'output',
            size_key: str = 'size',
    ) -> None:
        super().__init__(name, lambda x, y: x + y, value_key=value_key)
        self._size_key = size_key
        self._size = 0

    def _reset(self, state: dict) -> None:
        super()._reset(state)
        self._size = 0

    def _update(self, state: dict) -> None:
        super()._update(state)
        self._size += state.get(self._size_key, 1)

    def _compute(self, state: dict) -> None:
        super()._compute(state)
        state[self.name] /= self._size



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['rnnr']

package_data = \
{'': ['*']}

install_requires = \
['tqdm']

extras_require = \
{'dev': ['flake8', 'flake8-mypy', 'yapf'],
 'doc': ['Sphinx ~=1.8.5',
         'sphinx_rtd_theme',
         'sphinx-autodoc-typehints ~=1.6.0'],
 'test': ['pytest ~=4.3.0', 'pytest-cov']}

setup(name='rnnr',
      version='0.0.6',
      description='rnnr: neural network runner',
      author='Kemal Kurniawan',
      author_email='kemal@kkurniawan.com',
      url='https://github.com/kmkurn/rnnr',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6,<4',
     )



from unittest.mock import call, patch

import pytest

from rnnr.handlers import EarlyStopper


def test_ok(runner):
    # default patience is 5
    values = [5, 4, 3, 4, 5, 4, 3, 2, 1, 0, 1, 2, 3, 4, 5, 6, 7]
    min_values = [5, 4, 3, 3, 3, 3, 3, 2, 1, 0, 0, 0, 0, 0, 0, 0, 0]
    es = EarlyStopper()

    with patch.object(runner, 'stop', autospec=True) as mock_stop:
        for i, (v, mv) in enumerate(zip(values, min_values)):
            es({'runner': runner, 'loss': v})
            assert es.min_loss == pytest.approx(mv, abs=1e-4)
            if i == len(values) - 2:
                mock_stop.assert_called_once_with()
            elif i == len(values) - 1:
                assert mock_stop.mock_calls == [call(), call()]
            else:
                assert not mock_stop.called


def test_custom_patience(runner):
    values = [5, 4, 3, 4, 5, 6, 7, 8]
    es = EarlyStopper(patience=3)

    with patch.object(runner, 'stop', autospec=True) as mock_stop:
        for i, v in enumerate(values):
            es({'runner': runner, 'loss': v})
            if i == len(values) - 2:
                mock_stop.assert_called_once_with()
            elif i == len(values) - 1:
                assert mock_stop.mock_calls == [call(), call()]
            else:
                assert not mock_stop.called


def test_loss_key(runner):
    values = [5, 4, 3, 4, 5, 6, 7]
    es = EarlyStopper(patience=2, loss_key='foo')

    with patch.object(runner, 'stop', autospec=True) as mock_stop:
        for i, v in enumerate(values):
            es({'runner': runner, 'foo': v})
            if i == len(values) - 2:
                mock_stop.assert_called_once_with()
            elif i == len(values) - 1:
                assert mock_stop.mock_calls == [call(), call()]
            else:
                assert not mock_stop.called
